package cn.com.qmth.scancentral.service.impl;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.github.jeffreyning.mybatisplus.service.MppServiceImpl;
import com.qmth.boot.core.exception.ParameterException;
import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.core.fss.store.FileStore;
import com.qmth.boot.tools.uuid.FastUUID;

import cn.com.qmth.scancentral.bean.ImportSubjectDomain;
import cn.com.qmth.scancentral.bean.SubjectConfigDomain;
import cn.com.qmth.scancentral.bean.User;
import cn.com.qmth.scancentral.consumer.AnswerDataExportConsumer;
import cn.com.qmth.scancentral.dao.SubjectDao;
import cn.com.qmth.scancentral.entity.MarkSiteEntity;
import cn.com.qmth.scancentral.entity.SubjectEntity;
import cn.com.qmth.scancentral.enums.ExamStatus;
import cn.com.qmth.scancentral.service.BatchPaperService;
import cn.com.qmth.scancentral.service.DataExportTaskService;
import cn.com.qmth.scancentral.service.FileService;
import cn.com.qmth.scancentral.service.MarkSiteService;
import cn.com.qmth.scancentral.service.StudentService;
import cn.com.qmth.scancentral.service.SubjectService;
import cn.com.qmth.scancentral.support.SpringContextHolder;
import cn.com.qmth.scancentral.util.Calculator;
import cn.com.qmth.scancentral.util.FileUtil;
import cn.com.qmth.scancentral.vo.ExportCetMarkingQueryVo;
import cn.com.qmth.scancentral.vo.ExportCetVo;
import cn.com.qmth.scancentral.vo.ImportStudentQueryVo;
import cn.com.qmth.scancentral.vo.SubjectConfigVo;
import cn.com.qmth.scancentral.vo.examinfo.SubjectConfig;
import cn.com.qmth.scancentral.vo.subject.DataExportTaskVo;
import cn.com.qmth.scancentral.vo.subject.ScanProgressVo;
import cn.com.qmth.scancentral.vo.subject.SubjectBreachVo;
import cn.com.qmth.scancentral.vo.subject.SubjectCustStatusVo;
import cn.com.qmth.scancentral.vo.subject.SubjectScanProgressVo;
import cn.com.qmth.scancentral.vo.subject.TaskIdVo;

@Service
public class SubjectServiceImpl extends MppServiceImpl<SubjectDao, SubjectEntity> implements SubjectService {

    private static final Logger log = LoggerFactory.getLogger(SubjectService.class);

    private static ExecutorService exec;

    @Autowired
    private StudentService studentService;

    @Autowired
    private FileService fileService;

    @Autowired
    private FileStore fileStore;

    @Autowired
    private BatchPaperService batchPaperService;

    @Autowired
    private DataExportTaskService dataExportTaskService;

    @Autowired
    private MarkSiteService markSiteService;

    static {
        int threadCount = 5;
        exec = new ThreadPoolExecutor(threadCount, threadCount, 0L, TimeUnit.SECONDS,
                new LinkedBlockingQueue<>(threadCount * 2), r -> {
                    Thread t = new Thread(r);
                    return t;
                }, (r, executor) -> {
                    if (!executor.isShutdown()) {
                        try {
                            executor.getQueue().put(r);
                        } catch (InterruptedException e) {
                            throw new RuntimeException(e);
                        }
                    }
                });
    }

    @Transactional
    @Override
    public int cleanByExamId(Long examId, String subjectCode) {
        QueryWrapper<SubjectEntity> wrapper = new QueryWrapper<>();
        LambdaQueryWrapper<SubjectEntity> lw = wrapper.lambda();
        lw.eq(SubjectEntity::getExamId, examId);
        lw.eq(SubjectEntity::getCode, subjectCode);
        return this.baseMapper.delete(wrapper);
    }

    @Override
    public List<SubjectEntity> listByExamId(Long examId) {
        QueryWrapper<SubjectEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SubjectEntity::getExamId, examId);
        queryWrapper.lambda().orderByDesc(SubjectEntity::getCode);
        return this.list(queryWrapper);
    }

    @Override
    public SubjectEntity findByExamIdAndCode(Long examId, String code) {
        QueryWrapper<SubjectEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SubjectEntity::getExamId, examId);
        queryWrapper.lambda().eq(SubjectEntity::getCode, code);
        return this.getOne(queryWrapper);
    }

    @Override
    public int countByExamId(Long examId) {
        QueryWrapper<SubjectEntity> wrapper = new QueryWrapper<>();
        LambdaQueryWrapper<SubjectEntity> lw = wrapper.lambda();
        lw.eq(SubjectEntity::getExamId, examId);
        return this.count(wrapper);
    }

    @Transactional
    @Override
    public int importSubject(List<ImportSubjectDomain> subjects) {
        Map<Long, Set<String>> map = new HashMap<Long, Set<String>>();
        List<SubjectEntity> list = this.list();
        for (SubjectEntity subjectEntity : list) {
            Set<String> set = map.get(subjectEntity.getExamId());
            if (set == null) {
                set = new HashSet<String>();
            }
            set.add(subjectEntity.getCode());
            map.put(subjectEntity.getExamId(), set);
        }

        List<SubjectEntity> saveList = new ArrayList<SubjectEntity>();
        for (ImportSubjectDomain importSubjectDomain : subjects) {
            Set<String> set = map.get(importSubjectDomain.getExamId());
            if (set != null) {
                if (!set.contains(importSubjectDomain.getSubjectCode())) {
                    SubjectEntity subjectEntity = new SubjectEntity(importSubjectDomain.getExamId(),
                            importSubjectDomain.getSubjectCode(), importSubjectDomain.getSubjectName());
                    set.add(importSubjectDomain.getSubjectCode());
                    saveList.add(subjectEntity);
                }
            } else {
                set = new HashSet<String>();
                SubjectEntity subjectEntity = new SubjectEntity(importSubjectDomain.getExamId(),
                        importSubjectDomain.getSubjectCode(), importSubjectDomain.getSubjectName());
                set.add(importSubjectDomain.getSubjectCode());
                saveList.add(subjectEntity);
            }
            map.put(importSubjectDomain.getExamId(), set);
        }
        this.saveOrUpdateBatchByMultiId(saveList);
        return saveList.size();
    }

    @Transactional
    @Override
    public SubjectConfigVo config(User user, SubjectConfigDomain domain) {
        SubjectEntity s = this.findByExamIdAndCode(domain.getExamId(), domain.getSubjectCode());
        if (s == null) {
            throw new ParameterException("找不到对应的科目");
        }
        s.setPaperTypeBarcodeContent(domain.getPaperTypeBarcodeContent());
        this.saveOrUpdateByMultiId(s);
        return new SubjectConfigVo(s);
    }

    @Override
    public List<SubjectConfig> listConfigByExamId(Long examId) {
        QueryWrapper<SubjectEntity> queryWrapper = new QueryWrapper<>();
        queryWrapper.lambda().eq(SubjectEntity::getExamId, examId);
        queryWrapper.lambda().isNotNull(SubjectEntity::getPaperTypeBarcodeContent);
        List<SubjectEntity> list = this.list(queryWrapper);
        List<SubjectConfig> result = new ArrayList<SubjectConfig>();
        for (SubjectEntity subjectEntity : list) {
            result.add(new SubjectConfig(subjectEntity));
        }
        return result;
    }

    @Override
    public ScanProgressVo scanProgress(Long examId, String subjectCode) {
        ScanProgressVo ret = new ScanProgressVo();
        List<SubjectScanProgressVo> subjects = studentService.scanProgress(examId, subjectCode);
        ret.setSubjects(subjects);
        ret.setTotal(new SubjectScanProgressVo());
        if (CollectionUtils.isNotEmpty(subjects)) {
            Calendar now = Calendar.getInstance();
            long endTime = now.getTimeInMillis();
            now.add(Calendar.MINUTE, -1);
            long startTime = now.getTimeInMillis();
            for (SubjectScanProgressVo vo : subjects) {
                vo.setScannedCount(vo.getStudentCount() - vo.getUnexistCount());
                vo.setProgress(Calculator.percentage(vo.getScannedCount(), vo.getStudentCount(), 2));
                Integer count = batchPaperService.getScanStudentCount(examId, subjectCode, startTime, endTime);
                if (count == null || count == 0) {
                    vo.setEstimation("-");
                } else {
                    vo.setEstimation(Calculator.divide(
                            Calculator.multiply(Calculator.divide(count, endTime - startTime), vo.getUnexistCount()),
                            3600000) + "h");
                }
            }
            int totalUnScanned = studentService.getUnscannedCountByExam(examId);
            int totalCount = studentService.getCountByExam(examId);
            ret.getTotal().setStudentCount(totalCount);
            ret.getTotal().setUnexistCount(totalUnScanned);
            ret.getTotal().setProgress(
                    Calculator.percentage(ret.getTotal().getScannedCount(), ret.getTotal().getStudentCount(), 2));
            Integer count = batchPaperService.getScanStudentCount(examId, null, startTime, endTime);
            if (count == null || count == 0) {
                ret.getTotal().setEstimation("-");
            } else {
                ret.getTotal().setEstimation(
                        Calculator.divide(Calculator.multiply(Calculator.divide(count, endTime - startTime),
                                ret.getTotal().getUnexistCount()), 3600000) + "h");
            }
        }
        return ret;
    }

    @Override
    public List<SubjectBreachVo> breachList(Long examId) {
        List<SubjectBreachVo> ret = new ArrayList<>();
        List<SubjectEntity> subjects = listByExamId(examId);
        if (CollectionUtils.isNotEmpty(subjects)) {
            for (SubjectEntity e : subjects) {
                SubjectBreachVo vo = new SubjectBreachVo();
                ret.add(vo);
                vo.setSubjectCode(e.getCode());
                vo.setSubjectName(e.getName());
                String url = fileService.getBreachInfoUri(examId, e.getCode());
                try {
                    if (fileStore.exist(url)) {
                        vo.setBreachCount(Integer.valueOf(FileUtil.readFileContent(fileStore.read(url)).trim()));
                    } else {
                        vo.setBreachCount(0);
                    }
                } catch (Exception e1) {
                    throw new RuntimeException("文件访问异常", e1);
                }
            }
        }
        return ret;
    }

    @Override
    public List<SubjectCustStatusVo> custStatusList(Long examId) {
        List<SubjectCustStatusVo> ret = new ArrayList<>();
        List<SubjectEntity> subjects = listByExamId(examId);
        if (CollectionUtils.isNotEmpty(subjects)) {
            for (SubjectEntity e : subjects) {
                SubjectCustStatusVo vo = new SubjectCustStatusVo();
                ret.add(vo);
                vo.setSubjectCode(e.getCode());
                vo.setSubjectName(e.getName());
                String url = fileService.getCustStatusInfoUri(examId, e.getCode());
                try {
                    if (fileStore.exist(url)) {
                        vo.setCustStatusCount(Integer.valueOf(FileUtil.readFileContent(fileStore.read(url)).trim()));
                    } else {
                        vo.setCustStatusCount(0);
                    }
                } catch (Exception e1) {
                    throw new RuntimeException("文件访问异常", e1);
                }
            }
        }
        return ret;
    }

    @Override
    public TaskIdVo answerDataExport(Long examId, String subjectCode) {
        ImportStudentQueryVo sreq = new ImportStudentQueryVo();
        sreq.setExamId(examId);
        int scount = studentService.countByQuery(sreq);
        if (scount == 0) {
            throw new StatusException("没有考生信息");
        }
        DataExportTaskVo vo = new DataExportTaskVo();
        vo.setTotalCount(scount);
        vo.setExamId(examId);
        vo.setSubjectCode(subjectCode);
        vo.setSuccess(true);
        vo.setProgress(0.0);
        vo.setFileName("扫描答案.dbf");
        AnswerDataExportConsumer com = SpringContextHolder.getBean(AnswerDataExportConsumer.class);
        com.setVo(vo);
        exec.execute(com);
        dataExportTaskService.addTask(vo);
        return TaskIdVo.create(vo.getTaskId());
    }

    @Override
    public void answerDataExportDispose(DataExportTaskVo vo) {
        File[] files = null;
        try {
            File temDir = new File("temp/" + FastUUID.get() + "/");
            temDir.mkdirs();
            ExportCetMarkingQueryVo req = new ExportCetMarkingQueryVo();
            req.setExamId(vo.getExamId());
            req.setPageSize(100000);
            int pageNumber = 0;
            req.setPageNumber(pageNumber);
            Set<String> examNumbers = new HashSet<>();
            Map<String, MarkSiteEntity> cms = readCetMarking(vo.getExamId());
            while (true) {
                req.setPageNumber(++pageNumber);
                List<ExportCetVo> list = studentService.exportCetData(req);
                if (CollectionUtils.isEmpty(list)) {
                    break;
                }
                exportAnswer(examNumbers, temDir, list, cms, vo);
            }
            files = temDir.listFiles();
            if (files.length > 0) {
                File zip = new File(temDir.getAbsolutePath() + "/answer-data.zip");
                FileUtil.dozip(zip, files);
                vo.setFilePath(zip.getAbsolutePath());
            }
        } catch (StatusException e) {
            vo.setSuccess(false);
            vo.setErrMsg(e.getMessage());
        } catch (Exception e) {
            vo.setSuccess(false);
            vo.setErrMsg("系统异常");
            log.error("导出扫描答案DBF异常", e);
        } finally {
            if (files != null && files.length > 0) {
                for (File file : files) {
                    FileUtil.deleteDirectory(file);
                }
            }
        }
    }

    private void exportAnswer(Set<String> examNumbers, File temDir, List<ExportCetVo> list,
            Map<String, MarkSiteEntity> cms, DataExportTaskVo vo) {
        Map<String, List<String>> subjects = new HashMap<>();
        for (ExportCetVo data : list) {
            String key = data.getSubjectCode() + "-" + data.getExamNumber();
            if (examNumbers.contains(key)) {
                continue;
            }
            vo.setProgressCount(vo.getProgressCount() + 1);
            examNumbers.add(key);
            List<String> line = new ArrayList<>();
            line.add(data.getExamNumber());
            line.add(data.getName());
            if (ExamStatus.ABSENT.equals(data.getExamStatus())) {
                line.add("1");
            } else {
                line.add("0");
            }
            line.add(data.getSubjectCode());
            line.add(data.getAnswer());
            line.add(data.getCardFirst() + "");
            line.add(data.getCardSecond() + "");
            line.add(StringUtils.isBlank(data.getBreachCode()) ? "0" : data.getBreachCode());
            line.add(data.getPaperType());
            int subjectCode = Integer.valueOf(data.getExamNumber().substring(9, 10));
            String markingCode;
            if (subjectCode >= 3 && subjectCode <= 9) {
                markingCode = "88";
            } else {
                if (ExamStatus.ABSENT.equals(data.getExamStatus()) || "000000".equals(data.getPaperType())
                        || "999999".equals(data.getPaperType())) {
                    markingCode = "00";
                } else {
                    int site = Integer.valueOf(data.getExamNumber().substring(10, 13));
                    MarkSiteEntity cm = cms.get(data.getSubjectCode() + "-" + data.getPaperType());
                    if (cm == null) {
                        throw new StatusException(
                                "未找到评卷点数据,SubjectCode:" + data.getSubjectCode() + " PaperType:" + data.getPaperType());
                    }
                    if (site % 2 == 0) {
                        markingCode = cm.getEvenNumber();
                    } else {
                        markingCode = cm.getOddNumber();
                    }
                }
            }
            line.add(markingCode);
            List<String> lines = subjects.get(data.getSubjectCode());
            if (lines == null) {
                lines = new ArrayList<>();
                subjects.put(data.getSubjectCode(), lines);
            }
            lines.add(StringUtils.join(line, ","));
        }
        writeResult(subjects, temDir, "answer-data.txt");
    }

    private void writeResult(Map<String, List<String>> subjects, File rootDir, String fileName) {
        for (String k : subjects.keySet()) {
            File file = new File(rootDir.getAbsolutePath() + "/" + k + "/" + fileName);
            file.getParentFile().mkdirs();
            if (!file.exists()) {
                try {
                    file.createNewFile();
                } catch (IOException e) {
                    throw new StatusException("", e);
                }
            }
            try {
                FileUtils.writeLines(file, StandardCharsets.UTF_8.name(), subjects.get(k), true);
            } catch (IOException e) {
                throw new StatusException("文件处理出错", e);
            }
        }
    }

    private Map<String, MarkSiteEntity> readCetMarking(Long examId) {
        Map<String, MarkSiteEntity> ret = new HashMap<>();
        List<MarkSiteEntity> list = markSiteService.findByExam(examId);
        if (CollectionUtils.isNotEmpty(list)) {
            for (MarkSiteEntity e : list) {
                ret.put(e.getSubjectCode() + "-" + e.getPaperType(), e);
            }
        }
        return ret;
    }
}
